/* nag_lapacklin_dgesv (f07aac) Example Program.
 *
 * NLL6I27DB/NLL6I27DBL Version.
 *
 * Copyright 2019 Numerical Algorithms Group.
 *
 * Mark 27.0, 2019.
 */

#include <cassert>
#include <iostream>
#include <Eigen/Dense>
#include <stdio.h>
#include <nag.h>
#include <chrono>

using namespace std::chrono;

int main(int argc, char* argv[])
{
  assert(argc==3); Integer n=std::stoi(argv[1]), nrhs=std::stoi(argv[2]);

  Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> bb;
  Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> aa;
  /* this is the data to be fitted */
  aa = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(n,n);
  bb = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(n,nrhs);

  /* Scalars */
  Integer exit_status = 0, i, j, pda, pdb;
  double res = 0.0;
  /* Arrays */
  double *a = new double[n*n], *b = new double[n*nrhs];
  double *a_sav = new double[n*n], *b_sav = new double[n*nrhs];
  Integer *ipiv = new Integer[n];

  /* Nag Types */
  NagError fail;
  Nag_OrderType order;

  order = Nag_RowMajor;

  INIT_FAIL(fail);

  printf("nag_lapacklin_dgesv (f07aac) Example Program Results\n\n");

  pda = n;
  pdb = nrhs;


  /* Read A and B from Eigen Matrices */
  for (i = 0; i < n; ++i)
    for (j = 0; j < n; ++j)
      a[i*n+j] = aa(i,j);

  for (i = 0; i < n; ++i)
    for (j = 0; j < nrhs; ++j)
      b[i*nrhs+j] = bb(i,j);

  /* Save matrices for checking */
  for (i = 0; i<n*n; ++i) a_sav[i] = a[i];
  for (i = 0; i<n*nrhs; ++i) b_sav[i] = b[i];

  /* Solve the equations Ax = b for x using
   * nag_lapacklin_dgesv (f07aac)
   */
  auto start = high_resolution_clock::now();
  nag_lapacklin_dgesv(order, n, nrhs, a, pda, ipiv, b, pdb, &fail);
  if (fail.code != NE_NOERROR) {
    printf("Error from nag_lapacklin_dgesv (f07aac).\n%s\n", fail.message);
    exit_status = 1;
    return exit_status;
  }

  if ( n < 6 && nrhs < 6) {
    /* Print solution */
    printf("Solution\n");
    for (i = 0; i < n; ++i){
      for (j = 0; j < nrhs; ++j)
        printf("%11.4f ", b[i*nrhs+j]);
      printf("\n");
    }
  }

  /* Check the solution */
   nag_blast_dgemm(order, Nag_NoTrans, Nag_NoTrans, n, nrhs, n, 1.0, a_sav, pda,
            b, pdb, -1.0, b_sav, pdb, &fail);
   auto stop = high_resolution_clock::now();
   auto duration = duration_cast<milliseconds>(stop-start);
   std::cout << " Time Solve and Check  = " << duration.count() << "ms " << std::endl;

   res=0.0;
   for (i=0; i<n*nrhs; ++i){
     res += b_sav[i]*b_sav[i];
   }

   printf("\nResidual = %20.15f\n", res);

  delete [] a;
  delete [] b;
  delete [] ipiv;
  delete [] a_sav;
  delete [] b_sav;
  return exit_status;
}

#undef A
#undef B
